use sdl2::{event::Event, keyboard::Keycode, render::Canvas, video::Window};
use std::{time::Duration, time::Instant};

mod doodad;
mod inputstatus;
mod physics;
mod ship;
mod sprite;
mod system;

use crate::{
	doodad::Doodad, inputstatus::InputStatus, physics::Cartesian, physics::WorldPosition,
	ship::Ship, ship::ShipKind, sprite::SpriteAtlas, system::System,
};

trait Drawable {
	// Draw this item on the screen
	fn draw(&self, dc: &mut DrawContext, sa: &SpriteAtlas) -> Result<(), String>;
}

struct Camera {
	pos: Cartesian,
}

impl Camera {
	fn new() -> Self {
		Camera {
			pos: Cartesian::new(0.0, 0.0),
		}
	}
}

static FTL: f64 = 1.0 / 200.0; // frame time limit

struct DrawContext<'a> {
	canvas: &'a mut Canvas<Window>,
	camera: Camera,

	// Dimensions of the window
	window_size: Cartesian,

	// Position of the top-left corner of the window,
	// relative to camera position.
	top_left: Cartesian,
}

impl<'a> DrawContext<'a> {
	fn new(canvas: &'a mut Canvas<Window>, camera: Camera) -> Self {
		let mut s = Self {
			canvas,
			camera,
			window_size: Cartesian::new(0.0, 0.0),
			top_left: Cartesian::new(0.0, 0.0),
		};
		s.update();
		return s;
	}

	fn update(&mut self) {
		self.window_size = Cartesian::from(self.canvas.window().size());
		self.top_left = (self.window_size / 2.0) * Cartesian::new(-1.0, 1.0);
	}
}

fn main() -> Result<(), String> {
	let sdl_context = sdl2::init()?;
	let video_subsystem = sdl_context.video()?;

	let window = video_subsystem
		.window("SDL2", 640, 480)
		.position_centered()
		.resizable()
		.build()
		.map_err(|e| e.to_string())?;

	let mut canvas = window
		.into_canvas()
		.accelerated()
		.build()
		.map_err(|e| e.to_string())?;
	let texture_creator = canvas.texture_creator();

	let sa = SpriteAtlas::new(&texture_creator)?;
	let mut event_pump = sdl_context.event_pump()?;

	let mut system = System::new();

	let mut i = InputStatus::new();

	let mut s = Ship::new(ShipKind::Gypsum, Cartesian::new(0.0, 0.0));

	//let mut last_frame_time = 0f64;
	let mut frame_start;

	let mut running = true;

	let mut dc = DrawContext::new(&mut canvas, Camera::new());

	while running {
		frame_start = Instant::now();
		dc.update();

		for event in event_pump.poll_iter() {
			match event {
				Event::Quit { .. }
				| Event::KeyDown {
					keycode: Some(Keycode::Escape),
					..
				} => {
					running = false;
				}
				_ => i.update(event),
			}
		}

		dc.camera.pos = s.body.pos;

		// Draw
		dc.canvas.clear();
		system.draw(&mut dc, &sa)?;
		s.draw(&mut dc, &sa)?;
		dc.canvas.present();

		let frame_time = frame_start.elapsed().as_secs_f64();

		// Wait
		// (limit frame rate)
		if frame_time < FTL {
			std::thread::sleep(Duration::from_secs_f64(FTL - frame_time));
		}

		// Apply input

		let frame_time = frame_start.elapsed().as_secs_f64();
		//last_frame_time = frame_time;

		s.body.tick(frame_time);
		system.tick(frame_time);

		if i.key_thrust {
			s.body.thrust(50.0 * frame_time);
		}

		if i.key_right {
			s.body.rot(-1.0 * frame_time);
		}

		if i.key_left {
			s.body.rot(1.0 * frame_time);
		}
	}

	Ok(())
}
